/* Program Name:	Apri.c
   Author:		Jesus Gonzalez
   Date:		October 18, 1999
   Description:	This program was made to mine a database
			Using the Association Rules method.
   Class:		Data Mining, CSE6362
   Professor:	Diane J. Cook
   Homework:	2							*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "math.h"

#define MAXTRANSACTIONS 800	// Maximum number of transactions in the database
#define MAXFIELDS 50		// Maximum number of fields per transaction
#define MAXLINESIZE 200		// Maximum number of characters per transaction
#define MAXLABELSIZE 80		// Maximum number of characters per field
#define DIFFERENTLABELS 1000	// Maximum number of different labels
#define MAXELEMENTSSIZE 150	// Maximum number of characters per label
#define LSIZE 20			// Size of vector L[LSIZE] and C[LSIZE]
#define MINASUPPORT 6		// Minimum support used to mine association rules
#define MINACONFIDENCE 0.6	// Minimum confidence to mine association rules

/* To store the transactions */
struct _label_ {
  char label[MAXLABELSIZE];
  } transactions[MAXTRANSACTIONS][MAXFIELDS];

typedef struct _L_NODE_		// Type to store C[x] and L[x] information
{							// This type was made to be a list
  char elem[MAXELEMENTSSIZE];	// The label
  int  frequence;				// The frequency
  struct _L_NODE_ *next;		// Pointer to the next element in the list
} L_NODE, *PL_NODE;

typedef struct _RULE_		// Type to store the rules as a list
{
  char lhs[MAXELEMENTSSIZE];	// Left hand side of the rule
  char rhs[MAXELEMENTSSIZE];	// Right hand side of the rule
  double confidence;			// Confidence of the rule
  double support;				// Support of the rule
  struct _RULE_ *next;			// Next rule in the list
} R_NODE, *PR_NODE;

typedef struct _L_				// Type for the L and C lists
{
  PL_NODE head;
  PL_NODE tail;
  PL_NODE current;
  int numElements;
} L_LIST, *PL_LIST;

typedef struct _R_				// List for the rules R
{
  PR_NODE head;
  PR_NODE tail;
  PR_NODE current;
  int numElements;
} R_LIST, *PR_LIST;


// Global Variables

PL_LIST L;						// Pointer to the list L
PL_LIST C;						// Pointer to the list C
PR_LIST R;						// Pointer to the list R

int numFields;					// Number of fields in the database
int numTrans;					// Number of transactions in the database
int numLabels;					// Number of different labels in the whole database
char *tmp,*tmp1,*tmp2,*tmp3,*tmp4;  // Temporal variables


/* This function Creates and returns a new node to add to a C or L list */
PL_NODE CreateNode()
{
  PL_NODE newNode;

  newNode = (PL_NODE) malloc (sizeof(L_NODE));
  newNode->frequence = 0;
  strcpy(newNode->elem,"");
  return newNode;
}


/* This function Creates and returns a new node to add to the R rules list */
PR_NODE CreateRNode()
{
  PR_NODE newNode;

  newNode = (PR_NODE) malloc (sizeof(R_NODE));
  newNode->confidence = 0.0;
  newNode->support = 0.0;
  strcpy(newNode->lhs,"");
  strcpy(newNode->rhs,"");
  return newNode;
}


/* This function returns in tmp the element number "numElem" of the
   field elem of lset, the elements are separated by the character "," */
void GetElement(PL_NODE lset, int numElem, char *tmp)
{
  char elem[50];
  int cIndex = 0, eIndex = 0, rIndex = 0, numElemInSet = 0;


  if(strcmp(lset->elem,"")==0)
  {
	  strcpy(tmp,"");
	  return;
  }
  strcpy(elem,"");
  for(cIndex = 0; cIndex < (int) strlen(lset->elem); cIndex++)
  {
    if(lset->elem[cIndex] == ',')
      numElemInSet++;
  }  
  numElemInSet++;


  if(numElem > numElemInSet)
  {
    printf("Error, there are only %d elements in this set..., not %d\n",numElemInSet,numElem);
  return;
  }
  
  eIndex = 1;
  cIndex = 0;
  while(eIndex < numElem)
  {
    while(lset->elem[cIndex] != ',')
    {
      cIndex++;
    }
    cIndex++;
    eIndex++;
  }									
  while(lset->elem[cIndex] != ',' && cIndex <= (int) strlen(lset->elem))
  {
    elem[rIndex] = lset->elem[cIndex];
    rIndex++;
    cIndex++;
  }

  elem[rIndex] = '\0';
  strcpy(tmp,elem);
  return;
}


/* This function substracts from a the elements contained in b and
   returns the result in tmp. */
void Substract(char a[], char b[], char *tmp)
{
  int cIndex = 0, index1 = 0, index2 = 0, numElemInA = 0, numElemInB = 0, flag = 0;
  PL_NODE current;

  current = malloc(sizeof(L_NODE));

  strcpy(tmp,"");

  for(cIndex = 0; cIndex < (int) strlen(a); cIndex++)
  {
    if(a[cIndex] == ',')
      numElemInA++;
  }  
  numElemInA++;

  for(cIndex = 0; cIndex < (int) strlen(b); cIndex++)
  {
    if(b[cIndex] == ',')
      numElemInB++;
  }  
  numElemInB++;

  for (index1 = 1; index1 <= numElemInA; index1++)
  {
	flag = 0;
    for (index2 = 1; index2 <= numElemInB; index2++)
	{
		strcpy(current->elem,a);
		GetElement(current,index1,tmp2);
		strcpy(current->elem,b);
		GetElement(current,index2,tmp3);
		if(strcmp(tmp2,tmp3) == 0)
		{
			flag = 1;
		}
	}
	if (flag == 0)
	{
	  if(strcmp(tmp,"") == 0)
	  {
	    strcpy(tmp,tmp2);
	  }
	  else
	  {
	    strcat(tmp,",");
		strcat(tmp,tmp2);
	  }
	}
  }
  return;
}


/* This function prints the elements in the list L or C depending of
   the value of type, if type = 0 it works with list L, if type = 1
   it works with list C. Level is the index of the list as L[level]
   or C[level]. */
void PrintLabels(int level, int type)
{
  PL_NODE current;

  if (type == 0)
    current = L[level].head;
  else
    current = C[level].head;

  if(type == 0)
    if(current == NULL)
      printf("I'm sorry, this L list is empty, it has %d elements\n",L[level].numElements);
  if(type == 1)
    if(current == NULL)
      printf("I'm sorry, this C list is empty, it has %d elements\n",C[level].numElements);

  if(type == 1)
    printf("This level %d of C has %d elements...\n",level,C[level].numElements);

  while(current != NULL)
  {
    printf("Labels: %s Frequence: %d\n",current->elem,current->frequence);
    current = current->next;
  }
  return;
}


/* This function prints the rules in the list R, it prints the left
   hand side, then "===>", then the right hand side and finally the
   support and confidence of the rule. */
void PrintRules()
{
  PR_NODE current;
  int cIndex;
  int numElemInStr;
  int intField;
  int rulesCounter=1;
  char fieldName[50];
  PL_NODE currentTmp;

  currentTmp = CreateNode();
  current = R->head;

  if(current == NULL)
    printf("I'm sorry, there are no RULES, it has %d elements\n",R->numElements);

  while(current != NULL)
  {
	printf("\nRule %d: ",rulesCounter);
	rulesCounter++;
	numElemInStr = 0;
    for(cIndex = 0; cIndex < (int) strlen(current->lhs); cIndex++)
	{
      if (current->lhs[cIndex] == ',')
	    numElemInStr++;
	}
	numElemInStr++;

	for(cIndex = 0; cIndex < numElemInStr; cIndex++)
	{
	  strcpy(currentTmp->elem,current->lhs);
	  GetElement(currentTmp,cIndex+1,tmp2);
	  intField = atol(tmp2);
	  if (intField < 101)
	    strcpy(fieldName,"Origin");
	  else if (intField < 201)
	  {
		     intField = intField - 100;
			 strcpy(fieldName,"Cylinders");
	  }
			 else if (intField < 301)
			 {
			 intField = intField - 200;
			 strcpy(fieldName,"Mpg");
			 }
	  else if (intField < 401)
	  {
			 intField = intField - 300;
			 strcpy(fieldName,"Acceleration");
	  }
	  else if (intField < 501)
	  {
			 intField = intField - 400;
			 strcpy(fieldName,"Model year");
	  }
	  else if (intField < 801)
	  {
			 intField = intField - 500;
			 strcpy(fieldName,"Displacement");
	  }
	  else if (intField < 2001)
	  {
			 intField = intField - 800;
			 strcpy(fieldName,"Hosepower");
	  }
	  else
	  {
	    intField = intField - 2000;
		strcpy(fieldName,"Weight");
	  }
	  if(cIndex < numElemInStr-1)
	    printf("%s=%d and ",fieldName, intField);
	  else
	    printf("%s=%d ",fieldName, intField);
	}

    printf("=> ");

	numElemInStr = 0;
    for(cIndex = 0; cIndex < (int) strlen(current->rhs); cIndex++)
	{
      if (current->rhs[cIndex] == ',')
	    numElemInStr++;
	}
	numElemInStr++;

	for(cIndex = 0; cIndex < numElemInStr; cIndex++)
	{
	  strcpy(currentTmp->elem,current->rhs);
	  GetElement(currentTmp,cIndex+1,tmp2);
	  intField = atol(tmp2);
	  if (intField < 101)
	    strcpy(fieldName,"Origin");
	  else if (intField < 201)
	  {
		     intField = intField - 100;
			 strcpy(fieldName,"Cylinders");
	  }
			 else if (intField < 301)
			 {
			 intField = intField - 200;
			 strcpy(fieldName,"Mpg");
			 }
	  else if (intField < 401)
	  {
			 intField = intField - 300;
			 strcpy(fieldName,"Acceleration");
	  }
	  else if (intField < 501)
	  {
			 intField = intField - 400;
			 strcpy(fieldName,"Model year");
	  }
	  else if (intField < 801)
	  {
			 intField = intField - 500;
			 strcpy(fieldName,"Displacement");
	  }
	  else if (intField < 2001)
	  {
			 intField = intField - 800;
			 strcpy(fieldName,"Hosepower");
	  }
	  else
	  {
	    intField = intField - 2000;
		strcpy(fieldName,"Weight");
	  }
	  if(cIndex < numElemInStr-1)
	    printf("%s=%d and ",fieldName, intField);
	  else
	    printf("%s=%d ",fieldName, intField);
	}
	printf(" (confidence = %f, support = %f)\n",current->confidence,current->support);

    current = current->next;
  }
  return;
}


/* This function is used to determine if an element of the list C[level]
   will be prunned or not, 1 means that the element will be prunned and
   0 that the element will not be prunned.*/
int WillIPruneIt(char *elem, int level)
{
  PL_NODE current;
  int flag;  
  
  current = L[level-1].head;

  // 1 -> the element is not in L[level]
  // 0 -> the element is in L[level]
  flag = 1;
  while(current != NULL && flag == 1)
  {
    if(strcmp(elem,current->elem) == 0)
      flag = 0;
    current = current->next;
  }
  return flag;
}


/* This function prunes itemsets such that some (k-1)-subset of c is
   not in L */
void PruneC(int k)
{
  int index1, index2;
  PL_NODE current, previous;
  char *tmpElem;

  if( k == 2 )
  {
    return;
  }

  // get subsets to check, for k = 3 we have 3 subsets
  // 			   for k = 4 we have 4 subsets
  //			   for k = 5 we have 5 subsets
  tmpElem = malloc(MAXLINESIZE);

  current = C[k].head;
  previous = current;

  while (current != NULL)
  {
    for(index2 = 1; index2 <= k; index2++)
    {
      strcpy(tmpElem,"");
      for(index1 = 1; index1 <= k; index1++)
      {
        if(index1 != index2)
        {
	  GetElement(current, index1, tmp2);
	  if(strcmp(tmpElem, "") == 0)
	    strcat(tmpElem,tmp2);
	  else
	  {
	    strcat(tmpElem, ",");
	    strcat(tmpElem,tmp2);
	  }
        }
      }
      if(WillIPruneIt(tmpElem,k) == 1)	// 1 means yes, 0 means no
      {
	if(current == C[k].head)
	{
	  C[k].head = C[k].head->next;
	  free(current);
	  current = C[k].head;
	  previous = current;
	  C[k].numElements--;
	  break;
	}
	else if(current == C[k].tail)
	{
		previous->next = NULL;
		C[k].tail = previous;
		C[k].numElements--;
		free(current);
		current = NULL;
		break;
	}
	else
	{
	  previous->next = current->next;
	  free(current);
	  current = previous->next;
	  C[k].numElements--;
	  break;
	}
        break;
      }
    }
	if (current != NULL)
	{
	  previous = current;
	  current = current->next;
	}
  }
  free(tmpElem);
  return;
}


/* This function is used to calculate the frequence of the
   elements in C[level]. */
void CalculateFrequence(int level)
{
  PL_NODE current;
  int index1, index2, index3;
  int foundAll, foundThis;
  int mycounter=1;

  current = C[level].head;

  while( current != NULL )
  {
    mycounter++;
    current->frequence = 0;
    for(index1 = 0; index1 < numTrans - 1; index1++)
    {
      foundAll = 1;
      foundThis = 0;
      
      for(index2 = 0; index2 < level; index2++)
      {
        GetElement(current, index2+1, tmp2);
	foundThis = 0;
	for(index3 = 0; index3 < numFields; index3++)
	{
	  if(strcmp(tmp2,transactions[index1][index3].label) == 0)
	    foundThis = 1;
	}
	if(foundThis == 0)
	  foundAll = 0;
      }
      if(foundAll == 1)
	current->frequence++;
    }
    current = current->next;
  }
  return;
}


/* This function is used to calculate the support of a subset of
   elements in label, the number of elements is given by level. */
int GetSupport (char label[], int level)
{
  int support = 0;
  int index1, index2, index3;
  int foundAll, foundThis;
  PL_NODE current;

    current = CreateNode();
    support = 0;
    for(index1 = 0; index1 < numTrans-1; index1++)
    {
      foundAll = 1;
      foundThis = 0;
      
      for(index2 = 0; index2 < level; index2++)
      {
		strcpy(current->elem,label);
        GetElement(current, index2+1, tmp2);
	foundThis = 0;
	for(index3 = 0; index3 < numFields; index3++)
	{
	  if(strcmp(tmp2,transactions[index1][index3].label) == 0)
	    foundThis = 1;
	}
	if(foundThis == 0)
	  foundAll = 0;
      }
      if(foundAll == 1)
	support++;
    }
  free(current);
  return support;
}


/* This function is used to insert a label in C[level] or L[level]
   with frequence = ifrequence */
void InsertLabel(char label[], int level, int stype, int ifrequence)
{
  PL_NODE newNode;

  newNode = CreateNode();
  strcpy(newNode->elem,label);

  if(ifrequence == 0)
    newNode->frequence++;
  else
  if(ifrequence == -1)
    newNode->frequence = 0;
  else
    newNode->frequence = ifrequence;

  newNode->next = NULL;

  if(stype == 0)
  {
    if( L[level].head == NULL)
    {
      L[level].head = newNode;
      L[level].tail = newNode;
    }
    else
    {
      L[level].tail->next = newNode;
      L[level].tail = newNode;
    }
    numLabels++;
    L[level].numElements++;
  }
  else
  {
    if( C[level].head == NULL)
    {
      C[level].head = newNode;
      C[level].tail = newNode;
    }
    else
    {
      C[level].tail->next = newNode;
      C[level].tail = newNode;
    }

    numLabels++;
    C[level].numElements++;
  }
  return;
}


/* This function is used to insert a rule in the Rules list R */
void InsertRule(char lhs[], char rhs[], double confidence, double support)
{
  PR_NODE newNode;
  PR_NODE current;
  int found = 0;

  current = R->head;

  found = 0;
  while(current != NULL)
  {
    if(strcmp(current->lhs,lhs) == 0 && strcmp(current->rhs,rhs) == 0)
	  found = 1;
	current = current->next;
  }

  if (found == 0)
  {
    newNode = CreateRNode();
    strcpy(newNode->lhs,lhs);
    strcpy(newNode->rhs,rhs);
    newNode->confidence = confidence;
    newNode->support = support;

    newNode->next = NULL;

    if( R->head == NULL)
    {
      R->head = newNode;
      R->tail = newNode;
    }
    else
    {
      R->tail->next = newNode;
      R->tail = newNode;
    }

    R->numElements++;
  }
  return;
}


/* This function is used to create L[level] from C[level] when there
   is MINASUPPORT */
void CreateLk(int level)
{
  PL_NODE current;

  current = C[level].head;

  while (current != NULL)
  {
    if(current->frequence >= MINASUPPORT)
      InsertLabel(current->elem,level,(int) 0,current->frequence);
    current = current->next;
  }
  return;
}


/* This is the AprioriGen function make from the pseudocode, it is
   used to create C[k] from L[k-1]. */
void AprioriGen(int k)
{

  PL_NODE current1, current2;
  int eIndex, sameFlag;

  current1 = L[k-1].head;
  current2 = L[k-1].head->next;

  while(current1 != NULL)
  {
    current2 = current1->next;
  while(current2 != NULL)
  {
    sameFlag = 0;
    for(eIndex = 1; eIndex < k - 1; eIndex++)
    {
      GetElement(current1,eIndex,tmp2);
      GetElement(current2,eIndex,tmp3);
      if(strcmp(tmp2,tmp3) != 0)
	sameFlag = 1;
    }
    if(sameFlag == 0)
    {
      strcpy(tmp4,"");
      for(eIndex = 1; eIndex < k-1; eIndex++)
      {
        GetElement(current1,eIndex,tmp2);
        if(eIndex == 1)
          strcpy(tmp4,tmp2);
	else
	{
	  strcat(tmp4,",");
	  strcat(tmp4,tmp2);
	}
      }
      GetElement(current1,k-1,tmp2);
      GetElement(current2,k-1,tmp3);
      if(strcmp(tmp4,"") == 0)
      {
		  if(atoi(tmp2) < atoi(tmp3))
		  {
	strcpy(tmp4,tmp2);
	strcat(tmp4,",");
	strcat(tmp4,tmp3);
		  }
		  else
		  {
			strcpy(tmp4,tmp3);
			strcat(tmp4,",");
			strcat(tmp4,tmp2);
		  }
      } 
      else if(atoi(tmp2) < atoi(tmp3))
      {
        strcat(tmp4,",");
	strcat(tmp4,tmp2);
        strcat(tmp4,",");
	strcat(tmp4,tmp3);
      }
			else
			{
			  strcat(tmp4,",");
			  strcat(tmp4,tmp3);
			  strcat(tmp4,",");
			  strcat(tmp4,tmp2);
			}
    InsertLabel(tmp4,k,1,0);
    }
    current2 = current2->next;
  }
    current1 = current1->next;
  }
  PruneC(k);
  return;
}


/* This is the Apriori function from the pseudocode, used to create
   C[k] and from C[k], to create L[k] */
void Apriori()
{
  int k;

  k = 2;

  while( L[k-1].head != NULL )
  {
    AprioriGen(k);
    CalculateFrequence(k);
    CreateLk(k);
    k++;
  }
  return;
 }


/* This function deletes the return character from a string */
void delreturn(char * f)
{
  int flen,i;

  flen = 0;
  i = 0;

  flen = strlen(f);
  for (i=0;i<=flen;i++)
  {
    if(f[i] == '\n')
    {
      f[i] = '\0';
    }
  }
  return;
}


/* This function is used to find a substring in a string, if it is
   found, the function returns 0, otherwise it returns 1. */
int FindLabel(char label[])
{
  int LabelIndex = 0;
  int Found = 1;
  PL_NODE current;

  current = L[1].head;

  while( current != NULL && Found == 1)
  {
    if (strcmp(label,current->elem) == 0 )
      {
        Found = 0;
        current->frequence++;
      }
      current = current->next;
  }
  return Found;
}


/* This function is used to read the transactions and to store them in
   the transactions matrix. It also stores the different labels found
   in L[1] keeping track of its frequency. Finally it makes the labels
   different so that we can distinguish all the fields in the database,
   we do this by adding a number to each label depending of the field
   number. */
int ReadTransactions( char fileName[])
{
FILE *transFile;
char line[MAXLINESIZE];
char *tmpField, *tmpField2;
int transNumber;
int fieldNumber;
int numberOfFields = 0;
int i;
int addNumber = 0;
int tmpIntField;
int labelsCounter = 0;


  if ( ( transFile = fopen( fileName,"r" ) ) == NULL )
  {
    printf("Error at open file: %s\n", fileName);
    return -1;
  }
  else
  {
    tmpField = malloc (MAXLABELSIZE);
    tmpField2 = malloc (MAXLABELSIZE);
    strcpy(line,"");
    transNumber = 0;
    fieldNumber = 0;
    while ( fgets ( line, MAXLINESIZE, transFile ) != NULL )
    {
      numberOfFields = 0;
      for(i = 0; i < (int) strlen(line); i++)
      {
        if (line[i] == ' ')
          numberOfFields++;
      }
      numberOfFields++;

      tmpField = strtok(line, " " );
      while ( tmpField != NULL )
      {
        delreturn(tmpField);

	switch (fieldNumber)
	{
	  case 0:	addNumber = 0;
			break;
	  case 1:	addNumber = 100;
			break;
	  case 2:	addNumber = 200;
			break;
	  case 3:	addNumber = 300;
			break;
	  case 4:	addNumber = 400;
			break;
	  case 5:	addNumber = 500;
			break;
	  case 6:	addNumber = 800;
			break;
	  case 7:	addNumber = 2000;
			break;
	  default:	break;
	}
	
	tmpIntField = atol(tmpField);
        tmpIntField = tmpIntField + addNumber;
	sprintf(tmpField2,"%d",tmpIntField);

        strcpy( transactions[transNumber][fieldNumber].label, tmpField2 );
        if ( FindLabel(tmpField2) == 1 )
	{
          InsertLabel(tmpField2, 1, 0, 0);
	}
        fieldNumber++;
        tmpField = strtok(NULL, " ");
      }
      transNumber++;
      fieldNumber = 0;
    }
  }

  numFields = numberOfFields;
  numTrans = transNumber;
  free(tmpField);
  fclose(transFile);
  return 0;
}


/* This function is used to initialize L, we have a vector L[x] where
   x is of size LSIZE, then we have LSIZE levels for L and so we have
   space for LSIZE lists for L. */
void InitializeL()
{
  int lIndex = 0;

  L = (PL_LIST) malloc (LSIZE * sizeof(L_LIST));

  for(lIndex = 0; lIndex < LSIZE; lIndex++)
  {
    L[lIndex].head = NULL;
    L[lIndex].current = NULL;
    L[lIndex].tail = NULL;
    L[lIndex].numElements = 0;
  }
  return;
}


/* This function initializes C in the same way that we initialize L. */
void InitializeC()
{
  int lIndex = 0;

  C = (PL_LIST) malloc (LSIZE * sizeof(L_LIST));

  for(lIndex = 0; lIndex < LSIZE; lIndex++)
  {
    C[lIndex].head = NULL;
    C[lIndex].current = NULL;
    C[lIndex].tail = NULL;
    C[lIndex].numElements = 0;
  }
  return;
}


/* This function is used to initialize the list of Rules R. */
void InitializeR()
{
  int lIndex = 0;

  R = (PR_LIST) malloc (sizeof(R_LIST));
  R->head = NULL;
  R->current = NULL;
  R->tail = NULL;
  R->numElements = 0;

  return;
}


/* This function is used to generate the rules. First AssocRules sends
   and entry where lk = am, then GenRules calls itself recursively to
   try to generate rules with different combinations of elements in the
   lhs and rhs of the rule. The rule is generated only if there is
   enough confidence MINACONFIDENCE. */
void GenRules( char lk[], char am[] )
{
  int index1, index2, cIndex, numElemInAm, numElemInLk, numElemInSubAm;
  char *subAm, *subLk;
  double confidence = 0.0;
  double supportlk = 0.0, supportam = 0.0;
  PL_NODE current;

  current = CreateNode();
  subAm = malloc(sizeof(MAXLABELSIZE));
  subLk = malloc(sizeof(MAXLABELSIZE));

  numElemInAm = 0;
  numElemInLk = 0;
  numElemInSubAm = 0;

  // Calculating the number of elements of am.
  for(cIndex = 0; cIndex < (int) strlen(am); cIndex++)
  {
    if(am[cIndex] == ',')
      numElemInAm++;
  }  
  numElemInAm++;

  for(cIndex = 0; cIndex < (int) strlen(lk); cIndex++)
  {
    if(lk[cIndex] == ',')
      numElemInLk++;
  }  
  numElemInLk++;


  for (index1 = 1; index1 <= numElemInAm; index1++)
  {
	strcpy(subAm,"");
	strcpy(subLk,"");
	supportlk = 0.0;
	supportam = 0.0;
	confidence = 0.0;
    for (index2 = 1; index2 <= numElemInAm; index2++)
	{
      if (index2 != index1)
	  {
		strcpy(current->elem,am);
		GetElement(current,index2,tmp2);
        if (strcmp(subAm,"") == 0)
		{
		  strcpy(subAm,tmp2);
		}
		else
		{
		  strcat(subAm,",");
		  strcat(subAm,tmp2);
		}
	  }
	  else
	  {
		if (strcmp(subLk,"") == 0)
		{
		  strcpy(current->elem,am);
	      GetElement(current,index2,tmp2);
		  strcpy(subLk,tmp2);
		}
		else
		{
		  strcat(subLk,",");
		  strcat(subLk,tmp2);
		}
	  }
	}

	numElemInSubAm = 0;
    for(cIndex = 0; cIndex < (int) strlen(subAm); cIndex++)
	{
      if(subAm[cIndex] == ',')
      numElemInSubAm++;
	}  
    numElemInSubAm++;

	supportlk = (double) GetSupport(lk,numElemInLk);
	supportam = (double) GetSupport(subAm,numElemInSubAm);
	confidence = supportlk/supportam; 
	if (confidence >= MINACONFIDENCE)
	{
	  Substract(lk,subAm,tmp1);
      InsertRule(subAm,tmp1,confidence,supportlk);

	  if(numElemInAm - 1 > 1)
	  {
	    GenRules(lk,subAm);
	  }
	}
  }
  return;
}


/* This function is used to generate rules from all the elements of L
   in all its levels starting from level 2, then: L[2] - L[N], where
   L[N] is the last L that is not empty. */
void AssocRules()
{
  int index;
  PL_NODE current;

  index = 2;

  while (L[index].head != NULL)
  {
	current = L[index].head;
	while ( current != NULL)
	{
      GenRules(current->elem, current->elem);
	  current = current->next;
	}
	index++;
  }
  return;
}


/* This is our main program to generate association rules. */
int main (int argc, char *argv[])
{
int FStatus = 0;

  // Initializing variables
  tmp = malloc(MAXLINESIZE);
  tmp1 = malloc(MAXLINESIZE);
  tmp2 = malloc(MAXLINESIZE);
  tmp3 = malloc(MAXLINESIZE);
  tmp4 = malloc(MAXLINESIZE);
  strcpy(tmp,"");
  InitializeL();
  InitializeC();
  InitializeR();

  // Reading the transactions
  FStatus = ReadTransactions(argv[1]);

  // Calling the Apriori to generate Lk's
  Apriori();

  // Calling AssocRules to generate the rules from Lk's
  AssocRules();
  printf("                          ASSOCIATION RULES MINING \n");
  printf("_________________________________________________________________\n");
  printf("\n\n\n\n");
  printf("RULES............................................................\n");
  PrintRules();
}
