import java.util.*;
import java.io.*;


public class LargeItemset
	{
	int size; // value of k
	int minSupport = 6;
	int minConfidence = 60; 
	// Each element of the itemset will be a vector containing sorted
	// members of a set
	Vector itemset = new Vector(30, 10);
	/***************************************************************/
	LargeItemset(int x)
		{
		size = x;
		}
	/***************************************************************
	This function tries to find out whether the first k-1 items from 
    	item1 is equal to the first k-1 items of item2. If there is a
	match then it returns true otherwise false.
	****************************************************************/
	private	boolean	compatible(Item itm1, Item itm2)
		{
		Vector member1 = itm1.getMember();
		Vector member2	=	itm2.getMember();
		boolean joinable = true;
		String attribute1, attribute2;
		char column1, column2;
		int index1, index2, i;

		for (i = 0; i < member1.size() - 1; ++i)
			{
			attribute1 = (String) member1.elementAt(i);
			attribute2 = (String) member2.elementAt(i);
			if (attribute1.equals(attribute2) == false)
				joinable = false;
			}
		attribute1 = (String) member1.elementAt(i);
      attribute2 = (String) member2.elementAt(i);
      index1 = attribute1.indexOf('(');
      index2 = attribute2.indexOf('(');
      column1 = attribute1.charAt(index1 + 1);
      column2 = attribute2.charAt(index2 + 1);
		if (column1 == column2)
			joinable = false;
		return joinable;
		}
	/***************************************************************
	This function checks whether all elements of the small vector are
	present in the big vector or not
	****************************************************************/
	public boolean contains(Vector small, Vector big)
		{
		boolean result = true;

		for (int i = 0; i < small.size() && result == true; ++i)
			{
			String str = (String) small.elementAt(i);
			if (big.contains(str) == false)
				result = false;
			}
		return result;
		}
	/***************************************************************
	This function generates the rules from the large itemsets.
	****************************************************************/
	public void generateRules(int k, Vector largeSet, Database auto)
		{
      Vector rightSet = new Vector(8);
		if (k == 1) // rules of size 2
			{
			for (int i = 0; i < itemset.size(); ++i)
				{
				Item itm = (Item) itemset.elementAt(i);
				int support = itm.getSupport();
         	Vector member = itm.getMember();
				rightSet.addElement((String) member.elementAt(0));
				checkRule(support, member, rightSet, auto);
				rightSet.removeAllElements();
				}
			}
		else	// rules of size 3 or greater
			{
			LargeItemset prevL = (LargeItemset) largeSet.elementAt(k-1);
			Vector smallItemset = prevL.getItemset();
			for (int i = 0; i < itemset.size(); ++i)
				{
				Item bigItem = (Item) itemset.elementAt(i);
         	int support = bigItem.getSupport();
         	Vector member = bigItem.getMember();
				for (int j = 0; j < smallItemset.size(); ++j)
					{
					Item	smallItem = (Item) smallItemset.elementAt(j);
					Vector	smallMember = smallItem.getMember();
					if (contains(smallMember, member) == true) 
						checkRule(support, member, smallMember, auto);
					}
				}
			}
		}
	/***************************************************************/
	public	Vector	getItemset()
		{
		return 	itemset;
		}
	/***************************************************************/
	public int	getLength()
		{
		return itemset.size(); // length of a set in the large
					// itemset
		}
	/***************************************************************
	This function joins two sets of size k-1 in order to create a new
	set of size k in the large itemset.
	****************************************************************/
	private	Vector	joinItems(Item itm1, Item itm2)
		{
		Vector member1 = itm1.getMember();
      Vector member2 =  itm2.getMember();
		Vector joinedMember = new Vector(8, 1);
      String attribute1, attribute2;
		char column1, column2;
		int index1, index2, i;

      for (i = 0; i < member1.size() - 1; ++i)
         {
         attribute1 = (String) member1.elementAt(i);
			joinedMember.addElement(attribute1);
         }
		attribute1 = (String) member1.elementAt(i);
		attribute2 = (String) member2.elementAt(i);
		index1 = attribute1.indexOf('(');
		index2 = attribute2.indexOf('(');
		column1 = attribute1.charAt(index1 + 1);
		column2 = attribute2.charAt(index2 + 1);
		if (column1 < column2)
			{
			joinedMember.addElement(attribute1);
			joinedMember.addElement(attribute2);
			}
		else
			{
			joinedMember.addElement(attribute2);
         joinedMember.addElement(attribute1);
			}
      return joinedMember;
		}
	/***************************************************************
	This function populates a large itemset of size k taking help 
	from the large itemset of size k-1 (prevL)
	****************************************************************/
	public void populate (LargeItemset prevL, Database auto)
		{
		int count = 0;

		Vector	prevItemset = prevL.getItemset();
		if (size == 2)
			populateWithDouble(prevItemset, auto);
		else
			{
			Vector 	member1, member2;
			for (int i = 0; i < prevItemset.size() - 1; ++i)
				{
				Item itm1 = (Item)   prevItemset.elementAt(i);
				for (int j = i+1; j < prevItemset.size(); ++j)
					{
					Item itm2 = (Item) prevItemset.elementAt(j);				
					if (compatible(itm1, itm2) == true)
						{
						Vector temp = joinItems(itm1, itm2);
						count = auto.getCount(temp);
						if (count >= minSupport)
							{
							Item itemsetElement = new Item(temp, count);
                 	 	itemset.addElement(itemsetElement);
							}		
						}
					}
				}
			}
		}
	/***************************************************************
	This function creates a large itemset where k = 1
	****************************************************************/
	private void	populateWithDouble(Vector prevItemset, Database auto)
		{
		String attribute1, attribute2;
		char column1, column2;
		int index, count = 0;

		for (int i = 0; i < prevItemset.size() - 1; ++i)
			{
			Item itm1 = (Item)	prevItemset.elementAt(i); 
			attribute1 = itm1.getElementAt(0);
			index = attribute1.indexOf('(');
			column1 = attribute1.charAt(index+1);
			for (int j = i+1; j < prevItemset.size(); ++j)
				{
				Item itm2 = (Item) prevItemset.elementAt(j);
				attribute2 = itm2.getElementAt(0);
				index = attribute2.indexOf('(');
				column2 = attribute2.charAt(index+1);
				if (column1 != column2)
					{
					Vector 	temp = new Vector(2, 1);
					temp.addElement(attribute1);
					temp.addElement(attribute2);
				 	count = auto.getCount(temp);	
					if (count >= minSupport)
						{
						Item itemsetElement = new Item(temp, count);
						itemset.addElement(itemsetElement);
						}
					}
				} 
			} 
		} 
	/***************************************************************
	This function populates a large itemset where k = 1
	****************************************************************/
	public void populateWithSingle(Database auto)
		{
		int count;
		Vector singleElement = new Vector(300, 5);
		Vector distinctAccelerations = auto.getDistinctAccelerations();
		Vector distinctCylinders = auto.getDistinctCylinders();
		Vector distinctDisplacements = auto.getDistinctDisplacements();
		Vector distinctHorsePowers = auto.getDistinctHorsePowers();
		Vector distinctMpgs = auto.getDistinctMpgs();
		Vector distinctOrigins = auto.getDistinctOrigins(); 
		Vector distinctWeights = auto.getDistinctWeights();
		Vector distinctModelYears = auto.getDistinctModelYears(); 

		for (int i = 0; i < distinctAccelerations.size(); ++i)
			singleElement.addElement((String) 
					distinctAccelerations.elementAt(i));
		for (int i = 0; i < distinctCylinders.size(); ++i)
			singleElement.addElement((String) 
					distinctCylinders.elementAt(i));
		for (int i = 0; i < distinctDisplacements.size(); ++i)
			singleElement.addElement((String) 
					distinctDisplacements.elementAt(i));
		for (int i = 0; i < distinctHorsePowers.size(); ++i)
			singleElement.addElement((String) 
					distinctHorsePowers.elementAt(i));
		for (int i = 0; i < distinctMpgs.size(); ++i)
			singleElement.addElement((String) 
					distinctMpgs.elementAt(i));
		for (int i = 0; i < distinctOrigins.size(); ++i)
			singleElement.addElement((String) 
					distinctOrigins.elementAt(i));
		for (int i = 0; i < distinctWeights.size(); ++i)
			singleElement.addElement((String) 
					distinctWeights.elementAt(i));
		for (int i = 0; i < distinctModelYears.size(); ++i)
			singleElement.addElement((String) 
					distinctModelYears.elementAt(i));
		for (int i = 0; i < singleElement.size(); ++i)
			{ 
			Vector temp = new Vector(1, 1);
			temp.addElement((String) singleElement.elementAt(i));
			count = auto.getCount(temp);
			if (count >= minSupport)
				{
				Item itemsetElement = new Item((String)
						singleElement.elementAt(i), count); 
				itemset.addElement(itemsetElement);
				}
			}
		}
	/***************************************************************
	Prints all the sets inside a large itemset.
	****************************************************************/
	public 	void	print()
		{
		System.out.println("Large Itemset of size" + size + ":");
	 	for (int i = 0; i < itemset.size(); ++i)
			{
			Item itm = (Item) itemset.elementAt(i);
			itm.print();
			}		
		}
	/***************************************************************
	Prints the contents of a vector.
	****************************************************************/
	public 	void  printVector(Vector V)
		{
		System.out.println("Vector Content:");
		for (int i = 0; i < V.size(); ++i)
			System.out.print((String) V.elementAt(i));
		System.out.println();
		}	
	/***************************************************************
	Prepares the left hand side and right hand side of an association
	rule.
	****************************************************************/
	private	void	checkRule(int support, Vector member, Vector rightMember, 
												Database auto)
		{
		Vector leftMember = new Vector(4, 1);
		for (int i = 0; i < member.size(); ++i)
			{
			String attribute = (String) member.elementAt(i);
			if (rightMember.contains(attribute) == false)
				leftMember.addElement(attribute); 
			}
		printRule(support, leftMember, rightMember, auto);
		printRule(support, rightMember, leftMember, auto);
		}
	/***************************************************************
	Prints the attribute value with the name of the attribute
	****************************************************************/
	private	void 	printAttribute(String str)
		{
		String digits, type = " ";
		int index;
		char attribute;

		index = str.indexOf('(');
		digits = str.substring(0, index);
		attribute = str.charAt(index+1);
		switch(attribute)
			{
			case 'o' :	type = "Origin"; 
							break;
         case 'c' :	type = "Cylinder"; 
                     break;
         case 'm' : 	type = "Mpg"; 
                     break;
         case 'a' : 	type = "Acceleration"; 
                     break;
         case 'y' : 	type = "Model Year"; 
                     break;
         case 'd' : 	type = "Displacement"; 
                     break;
         case 'h' : 	type = "Horsepower"; 
                     break;
         case 'w' : 	type = "Weight"; 
                     break;
			}
		System.out.print(type+"="+digits);
		}
	/***************************************************************
	Finds out candidate rules based on their support and confidence
	in the database.
	****************************************************************/
	private	void 	printRule (int support, Vector leftMember, 
				Vector rightMember, Database auto) 
		{
		int count = auto.getCount(leftMember);
      int confidence = (support * 100)/count;

      if (confidence >= minConfidence)
         {
			System.out.println();
         printAttribute((String) leftMember.elementAt(0));
         for (int i = 1; i < leftMember.size(); ++i)
				{
            System.out.print(" AND "); 
				printAttribute((String) leftMember.elementAt(i));
				}
         System.out.print(" => ");
         printAttribute((String) rightMember.elementAt(0));
         for (int i = 1; i < rightMember.size(); ++i)
				{
            System.out.print(" AND ");
				printAttribute((String) rightMember.elementAt(i));
				}
			System.out.println();
         System.out.print(" (Confidence = " + confidence+", Support = ");
         System.out.println(support+")");
         }
		}
	} // class 
